/***********************************************************************\
*                               program.d                               *
*                                                                       *
*                            LOLA interpreter                           *
*                                                                       *
*                              Version 0.2                              *
*                                                                       *
*                         The program as a whole                        *
>***********************************************************************<
* Copyright (C) 2007 Stewart Gordon.                                    *
*                                                                       *
* This software is provided 'as-is', without any express or implied     *
* warranty.  In no event will the author be held liable for any damages *
* arising from the use of this software.                                *
*                                                                       *
* Permission is granted to anyone to use this software for any purpose, *
* including commercial applications, and to alter it and redistribute   *
* it freely, subject to the following restrictions:                     *
*                                                                       *
* - The origin of this software must not be misrepresented; you must    *
*   not claim that you wrote the original software.  If you use this    *
*   software in a product, an acknowledgment in the product             *
*   documentation would be appreciated but is not required.             *
* - Altered source versions must be plainly marked as such, and must    *
*   not be misrepresented as being the original software.               *
* - This notice may not be removed or altered from any source           *
*   distribution.                                                       *
*                                                                       *
* For documentation and other information related to this product,      *
* visit http://smjg.port5.com/wwwep/languages/lola/.                    *
* Email me with comments/suggestions/bug reports: smjg@iname.com        *
\***********************************************************************/
module smjg.prog.lola.program;
import
	smjg.prog.lola.exception,
	smjg.prog.lola.expression,
	smjg.prog.lola.integer,
	smjg.prog.lola.parser,
	smjg.libs.util.console;
debug import std.stdio;

class Program {
	static this() {
		eof = new Abstraction(new Abstraction(Variable(2)));
	}

	static Expression eof;

	Expression currentExpression;

	this(char[] code) {
		currentExpression = parse(code);
	}

	int run() {
		while (true) {
			debug {
				writefln("\nCurrent expression before WHNF: ");
				writefln(currentExpression);
				getch();
			}

			currentExpression = currentExpression.reduceToWHNF();

			debug {
				writefln("\nCurrent expression after WHNF: ");
				writefln(currentExpression);
				getch();
			}

			/*	To unchurch an integer:
			 *	c[+],[0],
			 *
			 *	Cases:
			 *	c is            result
			 *	{n}             [n]
			 *	\\b             [+]
			 *	\\\b			\[0]
			 *	In the last case, apply to something
			 */
			debug {
				Expression action = currentExpression.applyTo(Integer.C_ZERO);
				writefln(\n `Applied to \\a:`);
				writefln(action);
				getch();
				action = Integer.unchurch(action);
				writefln("\nUnchurched:");
				writefln(action);
				getch();

			} else {
				Expression action;
				try {
					action = Integer.unchurch(currentExpression.applyTo(Integer.C_ZERO));
				} catch (ProgramRuntimeException e) {
					throw new ProgramRuntimeException("Invalid operation");
				}
			}
			Expression cont = currentExpression.applyTo(Integer.C_ONE);
			Integer outChar = cast(Integer) action;

			if (outChar !is null && !outChar.church) {
				// a Church integer - output it
				debug {
					cOut.writefln("Output character: %s (%d)", cast(dchar) outChar.value, outChar.value);
				} else {
					cOut.write(cast(dchar) outChar.value);
				}
				currentExpression = cont;

			} else if (action is Increment(false)) {
				// \\b - generate exit code and then terminate
				try {
					action = Integer.unchurch(cont);
				} catch (ProgramRuntimeException e) {
					throw new ProgramRuntimeException("Exit code is not a Church integer");
				}

				Integer exitCode = cast(Integer) action;
				if (exitCode is null) {
					throw new ProgramRuntimeException(
						"Exit code is not a Church integer");
				}
				return cast(int) exitCode.value;

			} else {
				try {
					action = action.applyTo(Integer.ZERO).reduceToWHNF.deref;
				} catch (ProgramRuntimeException e) {
					throw new ProgramRuntimeException("Invalid operation");
				}

				if (action is Integer.ZERO) {
					// \\\b - get input
					dchar inChar = cIn.readChar;
					Expression arg;

					if (inChar == dchar.init) {
						arg = eof;
					} else {
						arg = Integer(inChar, true);
					}
					currentExpression = cont.applyTo(arg);
				} else {
					throw new ProgramRuntimeException("Invalid operation");
				}
			}
		}
	}
}
